@@ -29,6 +29,7 @@
 #include "Utils.hh"
 
 #include "XrdCl/XrdClFileSystem.hh"
+#include "XrdCl/XrdClCopyProcess.hh"
 
 namespace PyXRootD
 {
@@ -56,7 +57,14 @@ namespace PyXRootD
     copyprocess->AddJob( copyprocess, args, kwds );
     pystatus = copyprocess->Prepare( copyprocess, NULL, NULL );
     if ( !pystatus ) return NULL;
-    if ( PyDict_GetItemString( pystatus, "ok" ) == Py_False ) return pystatus;
+    if ( PyDict_GetItemString( pystatus, "ok" ) == Py_False )
+    {
+      PyObject *tuple = PyTuple_New(2);
+      PyTuple_SetItem(tuple, 0, pystatus);
+      Py_INCREF(Py_None);
+      PyTuple_SetItem(tuple, 1, Py_None);
+      return tuple;
+    }
 
     pystatus = copyprocess->Run( copyprocess, PyTuple_New(0), PyDict_New() );
     if ( !pystatus ) return NULL;
@@ -689,4 +697,36 @@ namespace PyXRootD
     bool status = self->filesystem->SetProperty( name, value );
     return status ? Py_True : Py_False;
   }
+
+  //----------------------------------------------------------------------------
+  //! Do a remote cat
+  //----------------------------------------------------------------------------
+  PyObject* FileSystem::Cat( FileSystem *self,
+                             PyObject   *args,
+                             PyObject   *kwds )
+  {
+    static const char *kwlist[] = { "source" , NULL };
+    const  char       *source = 0;
+
+    if ( !PyArg_ParseTupleAndKeywords( args, kwds, "s",
+         (char**) kwlist, &source ) ) Py_RETURN_NONE;
+
+    XrdCl::CopyProcess process;
+    XrdCl::PropertyList props, results;
+
+    props.Set( "source", source );
+    props.Set( "target", "stdio://-" );
+    props.Set( "dynamicSource", true );
+
+    XrdCl::XRootDStatus st = process.AddJob( props, &results );
+    if( !st.IsOK() )
+      return ConvertType<XrdCl::XRootDStatus>( &st );
+
+    st = process.Prepare();
+    if( !st.IsOK() )
+      return ConvertType<XrdCl::XRootDStatus>( &st );
+
+    st = process.Run( 0 );
+    return ConvertType<XrdCl::XRootDStatus>( &st );
+  }
 }
