@@ -42,6 +42,7 @@
 #include <sstream>
 #include <memory>
 #include <sys/time.h>
+#include <uuid/uuid.h>
 
 namespace
 {
@@ -460,6 +461,19 @@ namespace XrdCl
     }
 
     pFileUrl = new URL( url );
+
+    //--------------------------------------------------------------------------
+    // Add unique uuid to each open request so replays due to error/timeout
+    // recovery can be correctly handled.
+    //--------------------------------------------------------------------------
+    URL::ParamsMap cgi = pFileUrl->GetParams();
+    uuid_t uuid;
+    char requuid[37]= {0};
+    uuid_generate( uuid );
+    uuid_unparse( uuid, requuid );
+    cgi["xrdcl.requuid"] = requuid;
+    pFileUrl->SetParams( cgi );
+
     if( !pFileUrl->IsValid() )
     {
       log->Error( FileMsg, "[0x%x@%s] Trying to open invalid url: %s",
@@ -502,26 +516,6 @@ namespace XrdCl
     pOpenFlags = flags;
     OpenHandler *openHandler = new OpenHandler( this, handler );
 
-    bool redirect = pUseVirtRedirector && pFileUrl->IsMetalink();
-
-    //--------------------------------------------------------------------------
-    // If we don't want to redirect and it's a local file simply delgate
-    // to the LocalFileHandler
-    //--------------------------------------------------------------------------
-    if( pFileUrl->IsLocalFile() && !redirect )
-    {
-      Status st = pLFileHandler->Open( pFileUrl->GetURL().c_str(), flags,
-                                             mode, openHandler, timeout );
-      if( !st.IsOK() )
-      {
-        delete openHandler;
-        pStatus    = st;
-        pFileState = Error;
-      }
-
-      return st;
-    }
-
     Message           *msg;
     ClientOpenRequest *req;
     std::string        path = pFileUrl->GetPathWithFilteredParams();
@@ -538,13 +532,7 @@ namespace XrdCl
     params.followRedirects = pFollowRedirects;
     MessageUtils::ProcessSendParams( params );
 
-    Status st;
-    if( redirect )
-      st = MessageUtils::RedirectMessage( *pFileUrl, msg, openHandler,
-                                          params, pLFileHandler );
-    else
-      st = MessageUtils::SendMessage( *pFileUrl, msg, openHandler,
-                                      params, pLFileHandler );
+    Status st = IssueRequest( *pFileUrl, msg, openHandler, params );
 
     if( !st.IsOK() )
     {
@@ -603,12 +591,7 @@ namespace XrdCl
     params.stateful        = true;
     MessageUtils::ProcessSendParams( params );
 
-    Status st;
-    if( pDataServer->IsLocalFile() )
-      st = pLFileHandler->Close( closeHandler, timeout );
-    else
-      st = MessageUtils::SendMessage( *pDataServer, msg, closeHandler,
-                                      params, pLFileHandler );
+    Status st = IssueRequest( *pDataServer, msg, closeHandler, params );
 
     if( !st.IsOK() )
     {
@@ -677,12 +660,6 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->Stat( stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
-
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -724,15 +701,8 @@ namespace XrdCl
     params.stateful        = true;
     params.chunkList       = list;
     MessageUtils::ProcessSendParams( params );
-
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->Read( offset, size, buffer, stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
-
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -778,12 +748,6 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->Write( offset, size, buffer, stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
-
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -819,12 +783,6 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->Sync( stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
-
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -862,12 +820,6 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->Truncate( size, stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
-
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -942,12 +894,80 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
+    return SendOrQueue( *pDataServer, msg, stHandler, params );
+  }
+
+  //------------------------------------------------------------------------
+  // Write scattered data chunks in one operation - async
+  //------------------------------------------------------------------------
+  XRootDStatus FileStateHandler::VectorWrite( const ChunkList &chunks,
+                                              ResponseHandler *handler,
+                                              uint16_t         timeout )
+  {
+    //--------------------------------------------------------------------------
+    // Sanity check
+    //--------------------------------------------------------------------------
+    XrdSysMutexHelper scopedLock( pMutex );
+
+    if( pFileState != Opened && pFileState != Recovering )
+      return XRootDStatus( stError, errInvalidOp );
+
+    Log *log = DefaultEnv::GetLog();
+    log->Debug( FileMsg, "[0x%x@%s] Sending a vector write command for handle "
+                "0x%x to %s", this, pFileUrl->GetURL().c_str(),
+                *((uint32_t*)pFileHandle), pDataServer->GetHostId().c_str() );
+
+    //--------------------------------------------------------------------------
+    // Determine the size of the payload
+    //--------------------------------------------------------------------------
+
+    // the size of write vector
+    uint32_t payloadSize = sizeof(XrdProto::write_list) * chunks.size();
+
+    //--------------------------------------------------------------------------
+    // Build the message
+    //--------------------------------------------------------------------------
+    Message             *msg;
+    ClientWriteVRequest *req;
+    MessageUtils::CreateRequest( msg, req, payloadSize );
+
+    req->requestid = kXR_writev;
+    req->dlen      = sizeof(XrdProto::write_list) * chunks.size();
+
+    ChunkList *list   = new ChunkList();
+
+    //--------------------------------------------------------------------------
+    // Copy the chunk info
+    //--------------------------------------------------------------------------
+    XrdProto::write_list *writeList =
+        reinterpret_cast<XrdProto::write_list*>( msg->GetBuffer( 24 ) );
+
+
+
+    for( size_t i = 0; i < chunks.size(); ++i )
     {
-      XRootDStatus st = pLFileHandler->VectorRead( *list, buffer, stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
+      writeList[i].wlen   = chunks[i].length;
+      writeList[i].offset = chunks[i].offset;
+      memcpy( writeList[i].fhandle, pFileHandle, 4 );
+
+      list->push_back( ChunkInfo( chunks[i].offset,
+                                  chunks[i].length,
+                                  chunks[i].buffer ) );
     }
 
+    //--------------------------------------------------------------------------
+    // Send the message
+    //--------------------------------------------------------------------------
+    MessageSendParams params;
+    params.timeout         = timeout;
+    params.followRedirects = false;
+    params.stateful        = true;
+    params.chunkList       = list;
+    MessageUtils::ProcessSendParams( params );
+
+    XRootDTransport::SetDescription( msg );
+    StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
+
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -955,10 +975,10 @@ namespace XrdCl
   // Write scattered buffers in one operation - async
   //------------------------------------------------------------------------
   XRootDStatus FileStateHandler::WriteV( uint64_t            offset,
-                                              const struct iovec *iov,
-                                              int                 iovcnt,
-                                              ResponseHandler    *handler,
-                                              uint16_t            timeout )
+                                         const struct iovec *iov,
+                                         int                 iovcnt,
+                                         ResponseHandler    *handler,
+                                         uint16_t            timeout )
   {
     XrdSysMutexHelper scopedLock( pMutex );
 
@@ -1002,12 +1022,6 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->WriteV( offset, iov, iovcnt, stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
-
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -1048,12 +1062,6 @@ namespace XrdCl
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
 
-    if( pDataServer->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->Fcntl( arg, stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
-
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
 
@@ -1089,11 +1097,6 @@ namespace XrdCl
 
     XRootDTransport::SetDescription( msg );
     StatefulHandler *stHandler = new StatefulHandler( this, handler, msg, params );
-    if( pFileUrl->IsLocalFile() )
-    {
-      XRootDStatus st = pLFileHandler->Visa( stHandler, timeout );
-      return ExamineLocalResult( st, msg, stHandler );
-    }
 
     return SendOrQueue( *pDataServer, msg, stHandler, params );
   }
@@ -1322,10 +1325,11 @@ namespace XrdCl
     //--------------------------------------------------------------------------
     if( !status->IsOK() && status->code == errRedirect && pFollowRedirects )
     {
-      std::string root = "root", xroot = "xroot";
+      static const std::string root = "root", xroot = "xroot", file = "file";
       std::string msg = status->GetErrorMessage();
       if( !msg.compare( 0, root.size(), root ) ||
-          !msg.compare( 0, xroot.size(), xroot ) )
+          !msg.compare( 0, xroot.size(), xroot ) ||
+          !msg.compare( 0, file.size(), file ) )
       {
         OnStateRedirection( msg, message, userHandler, sendParams );
         return;
@@ -1356,9 +1360,12 @@ namespace XrdCl
       ClientRequest *req = (ClientRequest*)message->GetBuffer();
       switch( req->header.requestid )
       {
-        case kXR_read:  i.opCode = Monitor::ErrorInfo::ErrRead;  break;
-        case kXR_readv: i.opCode = Monitor::ErrorInfo::ErrReadV; break;
-        case kXR_write: i.opCode = Monitor::ErrorInfo::ErrWrite; break;
+        case kXR_read:   i.opCode = Monitor::ErrorInfo::ErrRead;  break;
+        case kXR_readv:  i.opCode = Monitor::ErrorInfo::ErrReadV; break;
+        case kXR_write:  i.opCode = Monitor::ErrorInfo::ErrWrite; break;
+        // TODO
+        // once we do major release we can replace this with 'ErrWriteV'
+        case kXR_writev: i.opCode = Monitor::ErrorInfo::ErrWrite; break;
         default: i.opCode = Monitor::ErrorInfo::ErrUnc;
       }
 
@@ -1473,11 +1480,11 @@ namespace XrdCl
       //------------------------------------------------------------------------
       case kXR_readv:
       {
-        ++pVCount;
+        ++pVRCount;
         size_t segs = req->header.dlen/sizeof(readahead_list);
         readahead_list *dataChunk = (readahead_list*)message->GetBuffer( 24 );
         for( size_t i = 0; i < segs; ++i )
-          pVBytes += dataChunk[i].rlen;
+          pVRBytes += dataChunk[i].rlen;
         pVSegs += segs;
         break;
       }
@@ -1491,6 +1498,20 @@ namespace XrdCl
         pWBytes += req->write.dlen;
         break;
       }
+
+      //------------------------------------------------------------------------
+      // Handle writev response
+      //------------------------------------------------------------------------
+      case kXR_writev:
+      {
+        ++pVWCount;
+        size_t size = req->header.dlen/sizeof(readahead_list);
+        XrdProto::write_list *wrtList =
+            reinterpret_cast<XrdProto::write_list*>( message->GetBuffer( 24 ) );
+        for( size_t i = 0; i < size; ++i )
+          pVWBytes += wrtList[i].wlen;
+        break;
+      }
     };
   }
 
@@ -1578,8 +1599,7 @@ namespace XrdCl
     if( pFileState == Opened )
     {
       msg->SetSessionId( pSessionId );
-      Status st = MessageUtils::SendMessage( *pDataServer, msg, handler,
-                                             sendParams, pLFileHandler );
+      Status st = IssueRequest( *pDataServer, msg, handler, sendParams );
 
       //------------------------------------------------------------------------
       // Invalid session id means that the connection has been broken while we
@@ -1710,8 +1730,7 @@ namespace XrdCl
 
     MessageUtils::ProcessSendParams( params );
 
-    return MessageUtils::SendMessage( *pDataServer, msg, handler,
-                                      params, pLFileHandler );
+    return IssueRequest( *pDataServer, msg, handler, params );
   }
 
   //----------------------------------------------------------------------------
@@ -1740,7 +1759,7 @@ namespace XrdCl
     ClientOpenRequest *req;
     URL u = url;
 
-    if( !url.GetPath().empty() )
+    if( url.GetPath().empty() )
       u.SetPath( pFileUrl->GetPath() );
 
     std::string path = u.GetPathWithFilteredParams();
@@ -1770,20 +1789,11 @@ namespace XrdCl
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
 
-    Status st;
     //--------------------------------------------------------------------------
-    // If a virtual redirector is in use, set it up in send parameters
-    // and redirect the message
+    // Issue the open request
     //--------------------------------------------------------------------------
-    if( pUseVirtRedirector && url.IsMetalink() )
-      st = MessageUtils::RedirectMessage( url, msg, openHandler,
-                                          params, pLFileHandler );
-    //--------------------------------------------------------------------------
-    // Otherwise do an ordinary send
-    //--------------------------------------------------------------------------
-    else
-      st = MessageUtils::SendMessage( url, msg, openHandler,
-                                      params, pLFileHandler );
+    Status st = IssueRequest( url, msg, openHandler, params );
+
     // if there was a problem destroy the open handler
     if( !st.IsOK() )
     {
@@ -1849,9 +1859,8 @@ namespace XrdCl
     {
       it->request->SetSessionId( pSessionId );
       ReWriteFileHandle( it->request );
-      Status st = MessageUtils::SendMessage( *pDataServer, it->request,
-                                             it->handler, it->params,
-                                             pLFileHandler );
+      Status st = IssueRequest( *pDataServer, it->request,
+                                 it->handler, it->params );
       if( !st.IsOK() )
         FailMessage( *it, st );
     }
@@ -1898,6 +1907,17 @@ namespace XrdCl
           memcpy( dataChunk[i].fhandle, pFileHandle, 4 );
         break;
       }
+      case kXR_writev:
+      {
+        ClientWriteVRequest *req =
+            reinterpret_cast<ClientWriteVRequest*>( msg->GetBuffer() );
+        XrdProto::write_list *wrtList =
+            reinterpret_cast<XrdProto::write_list*>( msg->GetBuffer( 24 ) );
+        size_t size = req->dlen / sizeof(XrdProto::write_list);
+        for( size_t i = 0; i < size; ++i )
+          memcpy( wrtList[i].fhandle, pFileHandle, 4 );
+        break;
+      }
     }
 
     Log *log = DefaultEnv::GetLog();
@@ -1920,25 +1940,33 @@ namespace XrdCl
       i.oTOD = pOpenTime;
       gettimeofday( &i.cTOD, 0 );
       i.rBytes = pRBytes;
-      i.vBytes = pVBytes;
-      i.wBytes = pWBytes;
-      i.vSegs  = pVSegs;
+      i.vBytes = pVRBytes;
+      i.wBytes = pWBytes + pVWBytes; //TODO once we can break ABI compatibility
+      i.vSegs  = pVSegs;             // we will add a special field for WriteV
       i.rCount = pRCount;
-      i.vCount = pVCount;
+      i.vCount = pVRCount;
       i.wCount = pWCount;
       i.status = status;
       mon->Event( Monitor::EvClose, &i );
     }
   }
 
-  XRootDStatus FileStateHandler::ExamineLocalResult( XRootDStatus &status,
-                                                     Message *msg,
-                                                     ResponseHandler *handler )
+  XRootDStatus FileStateHandler::IssueRequest( const URL         &url,
+                                               Message           *msg,
+                                               ResponseHandler   *handler,
+                                               MessageSendParams &sendParams )
   {
-    if( status.IsOK() )
-      pInTheFly.insert( msg );
-    else
-      delete handler;
-    return status;
+    // first handle Metalinks
+    if( pUseVirtRedirector && url.IsMetalink() )
+      return MessageUtils::RedirectMessage( url, msg, handler,
+                                            sendParams, pLFileHandler );
+
+    // than local file access
+    if( url.IsLocalFile() )
+      return pLFileHandler->ExecRequest( url, msg, handler, sendParams );
+
+    // and finally ordinary XRootD requests
+    return MessageUtils::SendMessage( url, msg, handler,
+                                      sendParams, pLFileHandler );
   }
 }
