@@ -34,13 +34,118 @@
 #include "XrdCl/XrdClForkHandler.hh"
 #include "XrdCl/XrdClPlugInInterface.hh"
 #include "XrdCl/XrdClPlugInManager.hh"
+#include "XrdCl/XrdClLocalFileTask.hh"
+#include "XrdCl/XrdClZipListHandler.hh"
 #include "XrdSys/XrdSysPthread.hh"
 
+#include <sys/stat.h>
+
 #include <memory>
+#include <algorithm>
+#include <iterator>
 
 namespace
 {
 
+  class LocalFS
+  {
+    public:
+
+      XrdCl::XRootDStatus Stat( const std::string       &path,
+                                XrdCl::ResponseHandler  *handler,
+                                uint16_t                 timeout )
+      {
+        using namespace XrdCl;
+
+        Log *log = DefaultEnv::GetLog();
+
+        struct stat ssp;
+        if( stat( path.c_str(), &ssp ) == -1 )
+        {
+          log->Error( FileMsg, "Stat: failed: %s", strerror( errno ) );
+          XRootDStatus *error = new XRootDStatus( stError, errErrorResponse,
+                                                  XProtocol::mapError( errno ),
+                                                  strerror( errno ) );
+          return QueueTask( error, 0, handler );
+        }
+
+        // TODO support other mode options
+        uint32_t flags = S_ISDIR( ssp.st_mode ) ? kXR_isDir : 0;
+
+        std::ostringstream data;
+        data << ssp.st_dev << " " << ssp.st_size << " " << flags << " "
+            << ssp.st_mtime;
+        log->Debug( FileMsg, data.str().c_str() );
+
+        StatInfo *statInfo = new StatInfo();
+        if( !statInfo->ParseServerResponse( data.str().c_str() ) )
+        {
+          log->Error( FileMsg, "Stat: ParseServerResponse failed." );
+          delete statInfo;
+          return QueueTask( new XRootDStatus( stError, errErrorResponse, kXR_FSError ),
+                            0, handler );
+        }
+
+        AnyObject *resp = new AnyObject();
+        resp->Set( statInfo );
+        return QueueTask( new XRootDStatus(), resp, handler );
+      }
+
+      static LocalFS& Instance()
+      {
+        static LocalFS instance;
+        return instance;
+      }
+
+    private:
+
+      //------------------------------------------------------------------------
+      // Private constructors
+      //------------------------------------------------------------------------
+      LocalFS() : jmngr( XrdCl::DefaultEnv::GetPostMaster()->GetJobManager() )
+      {
+
+      }
+
+      //------------------------------------------------------------------------
+      // Private copy constructors
+      //------------------------------------------------------------------------
+      LocalFS( const LocalFS& );
+
+      //------------------------------------------------------------------------
+      // Private assignment operator
+      //------------------------------------------------------------------------
+      LocalFS& operator=( const LocalFS& );
+
+      //------------------------------------------------------------------------
+      // QueueTask - queues error/success tasks for all operations.
+      // Must always return stOK.
+      // Is always creating the same HostList containing only localhost.
+      //------------------------------------------------------------------------
+      XrdCl::XRootDStatus QueueTask( XrdCl::XRootDStatus *st, XrdCl::AnyObject *resp,
+          XrdCl::ResponseHandler *handler )
+      {
+        using namespace XrdCl;
+
+        // if it is simply the sync handler we can release the semaphore
+        // and return there is no need to execute this in the thread-pool
+        SyncResponseHandler *syncHandler =
+            dynamic_cast<SyncResponseHandler*>( handler );
+        if( syncHandler )
+        {
+          syncHandler->HandleResponse( st, resp );
+          return XRootDStatus();
+        }
+
+        LocalFileTask *task = new LocalFileTask( st, resp, 0, handler );
+        jmngr->QueueJob( task );
+        return XRootDStatus();
+      }
+
+      XrdCl::JobManager *jmngr;
+
+  };
+
   //----------------------------------------------------------------------------
   // Get delimiter for the opaque info
   //----------------------------------------------------------------------------
@@ -242,9 +347,17 @@ namespace
           //--------------------------------------------------------------------
           if( it->IsManager() )
           {
+            ++pOutstanding;
             FileSystem *fs = new FileSystem( it->GetAddress() );
-            if( fs->Locate( pPath, pFlags, new DeallocFSHandler(fs, this), pExpires-::time(0)).IsOK() )
-              ++pOutstanding;
+            if( pOutstanding == 0 || // protect against overflow, short circuiting
+                                     // will make sure the other part won't be executed
+                !fs->Locate( pPath, pFlags, new DeallocFSHandler(fs, this),
+                             pExpires-::time(0)).IsOK() )
+            {
+              --pOutstanding;
+              pPartial = true;
+              delete fs;
+            }
           }
         }
 
@@ -314,7 +427,7 @@ namespace
       // Constructor
       //------------------------------------------------------------------------
       DirListStatHandler( XrdCl::DirectoryList *list,
-                          uint32_t                  index,
+                          uint32_t              index,
                           XrdCl::RequestSync   *sync ):
         pList( list ),
         pIndex( index ),
@@ -361,20 +474,38 @@ namespace
       RecursiveDirListCtx( const XrdCl::URL &url, const std::string &path,
                            XrdCl::DirListFlags::Flags flags,
                            XrdCl::ResponseHandler *handler, time_t expires ) :
-        pending( 1 ), dirList( new XrdCl::DirectoryList() ),
-        expires( expires ), handler( handler ), flags( flags ),
-        fs( new XrdCl::FileSystem( url ) )
+                             finalst( 0 ), pending( 1 ),
+                             dirList( new XrdCl::DirectoryList() ), expires( expires ),
+                             handler( handler ), flags( flags ),
+                             fs( new XrdCl::FileSystem( url ) )
       {
         dirList->SetParentName( path );
       }
 
       ~RecursiveDirListCtx()
       {
+        delete finalst;
         delete dirList;
         delete fs;
       }
 
-      XrdCl::XRootDStatus         status;
+      void UpdateStatus( const XrdCl::XRootDStatus &st )
+      {
+        using namespace XrdCl;
+
+        if( !finalst )
+        {
+          finalst = st.IsOK() ? new XRootDStatus() : new XRootDStatus( st );
+          return;
+        }
+
+        // if they disagree set the status to partial
+        if( ( finalst->IsOK() && !st.IsOK() ) ||
+            ( !finalst->IsOK() && st.IsOK() ) )
+          *finalst = XRootDStatus( stOK, suPartial );
+      }
+
+      XrdCl::XRootDStatus        *finalst;
       int                         pending;
       XrdCl::DirectoryList       *dirList;
       time_t                      expires;
@@ -384,22 +515,6 @@ namespace
       XrdSysMutex                 mtx;
   };
 
-  //----------------------------------------------------------------------------
-  // Exception for a recursive dirlist handler
-  //----------------------------------------------------------------------------
-  struct RecDirLsErr
-  {
-      RecDirLsErr() : status( 0 ) { }
-
-      RecDirLsErr( const XrdCl::XRootDStatus &st ) :
-        status( new XrdCl::XRootDStatus( st ) )
-      {
-
-      }
-
-      XrdCl::XRootDStatus *status;
-  };
-
   //----------------------------------------------------------------------------
   // Handle results for a recursive dirlist request
   //----------------------------------------------------------------------------
@@ -430,35 +545,24 @@ namespace
       {
         using namespace XrdCl;
 
+        Log *log = DefaultEnv::GetLog();
+        bool finalrsp = !( status->IsOK() && status->code == XrdCl::suContinue );
         XrdSysMutexHelper scoped( pCtx->mtx );
 
-        try
-        {
-          // decrement the number of pending DieLists
+        // check if we have to continue with the same handler (the response
+        // has been chunked), if not we can decrement the number of pending
+        // DieLists
+        if( finalrsp )
           --pCtx->pending;
 
-          // check if the job hasn't failed somewhere else,
-          // if yes we just give up
-          if( !pCtx->status.IsOK() )
-            throw RecDirLsErr();
-
-          // check if we failed, if yes update the global status
-          // for all call-backs and call the user handler
-          if( !status->IsOK() )
-            throw RecDirLsErr( *status );
-
-          // check if we got a response ...
-          if( !response )
-            throw RecDirLsErr( XRootDStatus( stError, errInternal ) );
+        pCtx->UpdateStatus( *status );
 
+        if( status->IsOK() )
+        {
           // get the response
           DirectoryList *dirList = 0;
           response->Get( dirList );
 
-          // check if the response is not empty ...
-          if( !dirList )
-            throw RecDirLsErr( XRootDStatus( stError, errInternal ) );
-
           std::string parent = pCtx->dirList->GetParentName();
 
           DirectoryList::Iterator itr;
@@ -467,13 +571,15 @@ namespace
             DirectoryList::ListEntry *entry = *itr;
             StatInfo *info = entry->GetStatInfo();
             if( !info )
-              throw RecDirLsErr( XRootDStatus( stError, errNotSupported ) );
+            {
+              log->Error( FileMsg, "Recursive directory list operation for %s failed: "
+                          "kXR_dirlist with stat operation not supported.",
+                          parent.c_str() );
+              pCtx->UpdateStatus( XRootDStatus( stError, errNotSupported ) );
+              continue;
+            }
             std::string path = dirList->GetParentName() + entry->GetName();
 
-            // check the prefix
-            if( path.find( parent ) != 0 )
-              throw RecDirLsErr( XRootDStatus( stError, errInternal ) );
-
             // add new entry to the result
             path = path.substr( parent.size() );
             entry->SetStatInfo( 0 ); // StatInfo is no longer owned by dirList
@@ -499,45 +605,59 @@ namespace
               {
                 timeout = pCtx->expires - ::time( 0 );
                 if( timeout <= 0 )
-                  throw RecDirLsErr( XRootDStatus( stError, errOperationExpired ) );
+                {
+                  log->Error( FileMsg, "Recursive directory list operation for %s expired.",
+                              parent.c_str() );
+                  pCtx->UpdateStatus( XRootDStatus( stError, errOperationExpired ) );
+                  break;
+                }
               }
               // send the request
-              XRootDStatus st = pCtx->fs->DirList( parent + path, flags, handler, timeout );
+              std::string child = parent + path;
+              XRootDStatus st = pCtx->fs->DirList( child, flags, handler, timeout );
               if( !st.IsOK() )
-                throw RecDirLsErr( st );
+              {
+                log->Error( FileMsg, "Recursive directory list operation for %s failed: ",
+                            child.c_str(), st.ToString().c_str() );
+                pCtx->UpdateStatus( st );
+                continue;
+              }
             }
           }
-
-          if( pCtx->pending == 0 )
-          {
-            AnyObject *resp = new AnyObject();
-            resp->Set( pCtx->dirList );
-            pCtx->dirList = 0; // dirList is no longer owned by pCtx
-            pCtx->handler->HandleResponse( new XRootDStatus(), resp );
-          }
         }
-        catch( RecDirLsErr &ex )
+
+        // if there are no more outstanding dirlist queries we can finalize the request
+        if( pCtx->pending == 0 )
         {
-          if( ex.status )
-          {
-            pCtx->status = *ex.status;
-            pCtx->handler->HandleResponse( ex.status, 0 );
-          }
+          AnyObject *resp = new AnyObject();
+          resp->Set( pCtx->dirList );
+          pCtx->dirList = 0; // dirList is no longer owned by pCtx
+          pCtx->handler->HandleResponse( pCtx->finalst, resp );
+          pCtx->finalst = 0; // status is no longer owned by pCtx
+
+          // finalize the common context
+          scoped.UnLock();
+          delete pCtx;
+        }
+        // if the user requested chunked response we give what we have to the user handler
+        else if( status->IsOK() && ( pCtx->flags & DirListFlags::Chunked ) )
+        {
+          std::string parent = pCtx->dirList->GetParentName();
+          AnyObject *resp = new AnyObject();
+          resp->Set( pCtx->dirList );
+          pCtx->dirList = new XrdCl::DirectoryList();
+          pCtx->dirList->SetParentName( parent );
+          pCtx->handler->HandleResponse( new XRootDStatus( stOK, suContinue ), resp );
         }
 
-        // clean up the context if necessary
-        bool delctx = ( pCtx->pending == 0 );
-        scoped.UnLock();
-        if( delctx )
-          delete pCtx;
         // clean up the arguments
         delete status;
         delete response;
-        // and finally commit suicide
-        delete this;
+        // if we wont be continuing with the same handler, it can be deleted
+        if( finalrsp )
+          delete this;
       }
 
-
     private:
 
       RecursiveDirListCtx *pCtx;
@@ -574,7 +694,8 @@ namespace
   {
     public:
 
-      MergeDirListHandler( XrdCl::ResponseHandler *handler ) : pHandler( handler )
+      MergeDirListHandler( bool allowChunked, XrdCl::ResponseHandler *handler ) :
+        allowChunked( allowChunked ), pHandler( handler )
       {
 
       }
@@ -582,9 +703,13 @@ namespace
       virtual void HandleResponse( XrdCl::XRootDStatus *status,
                                    XrdCl::AnyObject    *response )
       {
+        XrdSysMutexHelper lck( mtx );
+
+        bool finalrsp = !( status->IsOK() && status->code == XrdCl::suContinue );
+
         try
         {
-          if( status->IsOK() )
+          if( !status->IsOK() )
             throw MergeDirLsErr( status, response );
 
           if( !response )
@@ -596,7 +721,11 @@ namespace
           if( !dirlist )
             throw MergeDirLsErr();
 
-          Merge( dirlist );
+          if( allowChunked )
+            MergeChunked( dirlist );
+          else
+            Merge( dirlist );
+
           response->Set( dirlist );
           pHandler->HandleResponse( status, response );
         }
@@ -606,7 +735,51 @@ namespace
           pHandler->HandleResponse( err.status, err.response );
         }
 
-        delete this;
+        if( finalrsp )
+        {
+          lck.UnLock();
+          delete this;
+        }
+      }
+
+      void MergeChunked( XrdCl::DirectoryList *&response )
+      {
+        using namespace XrdCl;
+
+        std::set<ListEntry*, less> unique;
+        // set of unique list entries from the response
+        std::set<ListEntry*, less> tmp( response->Begin(), response->End() );
+        // all the unique list entries that were not reported so far
+        std::set_difference( tmp.begin(),         tmp.end(),
+                             uniquesofar.begin(), uniquesofar.end(),
+                             std::inserter( unique, unique.end() ) );
+
+        // we update the set of unique list entries that were already
+        // reported to the user's handler
+        for( auto itr = unique.begin(); itr != unique.end(); ++itr )
+        {
+          ListEntry *ent = *itr;
+          if( !uniquesofar.count( ent ) )
+          {
+            StatInfo  *info   = ent->GetStatInfo() ? new StatInfo( *ent->GetStatInfo() ) : 0;
+            ListEntry *newent = new ListEntry( ent->GetHostAddress(), ent->GetName(), info );
+            uniquesofar.insert( newent );
+          }
+        }
+
+        DirectoryList *dirlist = new DirectoryList();
+        dirlist->SetParentName( response->GetParentName() );
+        for( auto itr = unique.begin(); itr != unique.end(); ++itr )
+        {
+          ListEntry *entry = *itr;
+          dirlist->Add( new ListEntry( entry->GetHostAddress(),
+                                       entry->GetName(),
+                                       entry->GetStatInfo() ) );
+          entry->SetStatInfo( 0 );
+        }
+
+        delete response;
+        response = dirlist;
       }
 
       static void Merge( XrdCl::DirectoryList *&response )
@@ -626,7 +799,7 @@ namespace
 
         delete response;
         response = dirlist;
-      }
+}
 
     private:
 
@@ -661,7 +834,10 @@ namespace
         }
       };
 
-      XrdCl::ResponseHandler *pHandler;
+      bool                        allowChunked;
+      XrdSysMutex                 mtx;
+      std::set<ListEntry*, less>  uniquesofar;
+      XrdCl::ResponseHandler     *pHandler;
   };
 }
 
@@ -692,14 +868,17 @@ namespace XrdCl
         {
           HostList::reverse_iterator it;
           for( it = hostList->rbegin(); it != hostList->rend(); ++it )
-          if( it->loadBalancer )
-          {
-            pFS->AssignLoadBalancer( it->url );
-            break;
-          }
+            if( it->loadBalancer )
+            {
+              pFS->AssignLoadBalancer( it->url );
+              break;
+            }
         }
+
+        bool finalrsp = !( status->IsOK() && status->code == suContinue );
         pUserHandler->HandleResponseWithHosts( status, response, hostList );
-        delete this;
+        if( finalrsp )
+          delete this;
       }
 
     private:
@@ -1176,6 +1355,9 @@ namespace XrdCl
     if( pPlugIn )
       return pPlugIn->Stat( path, handler, timeout );
 
+    if( pUrl->IsLocalFile() )
+      return LocalFS::Instance().Stat( path, handler, timeout );
+
     std::string fPath = FilterXrdClCgi( path );
 
     Message           *msg;
@@ -1300,6 +1482,17 @@ namespace XrdCl
     URL url = URL( path );
     std::string fPath = FilterXrdClCgi( path );
 
+    if( flags & DirListFlags::Zip )
+    {
+      // stat the file to check if it is a directory or a file
+      // the ZIP handler will take care of the rest
+      ZipListHandler *zipHandler = new ZipListHandler( *pUrl, path, flags, handler, timeout );
+      XRootDStatus st = Stat( path, zipHandler, timeout );
+      if( !st.IsOK() )
+        delete zipHandler;
+      return st;
+    }
+
     Message           *msg;
     ClientDirlistRequest *req;
     MessageUtils::CreateRequest( msg, req, fPath.length() );
@@ -1314,10 +1507,12 @@ namespace XrdCl
       handler = new RecursiveDirListHandler( *pUrl, url.GetPath(), flags, handler, timeout );
 
     if( flags & DirListFlags::Merge )
-      handler = new MergeDirListHandler( handler );
+      handler = new MergeDirListHandler( flags & DirListFlags::Chunked, handler );
 
     msg->Append( fPath.c_str(), fPath.length(), 24 );
     MessageSendParams params; params.timeout = timeout;
+    if( flags & DirListFlags::Chunked )
+      params.chunkedResponse = true;
     MessageUtils::ProcessSendParams( params );
     XRootDTransport::SetDescription( msg );
 
@@ -1332,6 +1527,20 @@ namespace XrdCl
                                     DirectoryList       *&response,
                                     uint16_t              timeout )
   {
+    //--------------------------------------------------------------------------
+    // Chunked response is only possible for async DirList call
+    //--------------------------------------------------------------------------
+    if( flags & DirListFlags::Chunked )
+      return XRootDStatus( stError, errNotSupported );
+
+    //--------------------------------------------------------------------------
+    // If the path ends with '.zip' extension add Zip flag
+    //--------------------------------------------------------------------------
+    static const std::string zip_sufix = ".zip";
+    if( path.size() >= zip_sufix.size() &&
+        std::equal( zip_sufix.rbegin(), zip_sufix.rend(), path.rbegin() ) )
+      flags |= DirListFlags::Zip;
+
     //--------------------------------------------------------------------------
     // We do the deep locate and ask all the returned servers for the list
     //--------------------------------------------------------------------------
@@ -1342,7 +1551,8 @@ namespace XrdCl
       //------------------------------------------------------------------------
       LocationInfo *locations;
       std::string locatePath = "*"; locatePath += path;
-      XRootDStatus st = DeepLocate( locatePath, OpenFlags::PrefName, locations );
+      XRootDStatus st = DeepLocate( locatePath,
+          OpenFlags::PrefName | OpenFlags::Compress | OpenFlags::IntentDirList, locations );
 
       if( !st.IsOK() )
         return st;
@@ -1520,7 +1730,8 @@ namespace XrdCl
     MessageUtils::CreateRequest( msg, req, list.length() );
 
     req->requestid  = kXR_prepare;
-    req->options    = flags;
+    req->options    = 0xff & flags;
+    req->optionX    = 0xffff & ( flags >> 8 );
     req->prty       = priority;
     req->dlen       = list.length();
 
