@@ -83,7 +83,7 @@ namespace XrdCl
       action  = handler->Examine( msg );
 
       if( action & IncomingMsgHandler::RemoveHandler )
-	pHandlers.erase( it );
+        pHandlers.erase( it );
     }
 
     if( !(action & IncomingMsgHandler::Take) )
@@ -113,10 +113,10 @@ namespace XrdCl
 
       if( action & IncomingMsgHandler::Take )
       {
-	if( !(action & IncomingMsgHandler::NoProcess ) )
-	  handler->Process( it->second );
+        if( !(action & IncomingMsgHandler::NoProcess ) )
+          handler->Process( it->second );
 
-	pMessages.erase( it );
+        pMessages.erase( it );
       }
     }
 
@@ -151,8 +151,8 @@ namespace XrdCl
       act     = handler->Examine( msg );
       exp     = it->second.second;
 
-      if( act & IncomingMsgHandler::Take )
-	pHandlers.erase( it );
+      if( act & IncomingMsgHandler::RemoveHandler )
+        pHandlers.erase( it );
     }
 
     if( handler )
@@ -199,9 +199,13 @@ namespace XrdCl
       action = it->second.first->OnStreamEvent( event, streamNum, status );
 
       if( action & IncomingMsgHandler::RemoveHandler )
-	pHandlers.erase( it++ );
+      {
+        auto next = it; ++next;
+        pHandlers.erase( it );
+        it = next;
+      }
       else
-	++it;
+        ++it;
     }
   }
 
@@ -219,12 +223,15 @@ namespace XrdCl
     {
       if( it->second.second <= now )
       {
-	it->second.first->OnStreamEvent( IncomingMsgHandler::Timeout, 0,
-				  Status( stError, errOperationExpired ) );
-	pHandlers.erase( it++ );
+        uint8_t act = it->second.first->OnStreamEvent( IncomingMsgHandler::Timeout, 0,
+                                         Status( stError, errOperationExpired ) );
+        auto next = it; ++next;
+        if( act & IncomingMsgHandler::RemoveHandler )
+          pHandlers.erase( it );
+        it = next;
       }
       else
-	++it;
+        ++it;
     }
   }
 }
