@@ -32,7 +32,6 @@
 #include "XrdOuc/XrdOucGMap.hh"
 #include "XrdSys/XrdSysTimer.hh"
 #include "XrdOuc/XrdOucPinLoader.hh"
-
 #include "XrdHttpTrace.hh"
 #include "XrdHttpProtocol.hh"
 //#include "XrdXrootd/XrdXrootdStats.hh"
@@ -52,7 +51,6 @@
 #define XRHTTP_TK_GRACETIME     600
 
 
-
 /******************************************************************************/
 /*                               G l o b a l s                                */
 /******************************************************************************/
@@ -74,6 +72,7 @@ int XrdHttpProtocol::Window = 0;
 char *XrdHttpProtocol::sslcert = 0;
 char *XrdHttpProtocol::sslkey = 0;
 char *XrdHttpProtocol::sslcadir = 0;
+char *XrdHttpProtocol::sslcipherfilter = 0;
 char *XrdHttpProtocol::listredir = 0;
 bool XrdHttpProtocol::listdeny = false;
 bool XrdHttpProtocol::embeddedstatic = true;
@@ -88,12 +87,15 @@ char *XrdHttpProtocol::secretkey = 0;
 
 char *XrdHttpProtocol::gridmap = 0;
 XrdOucGMap *XrdHttpProtocol::servGMap = 0;  // Grid mapping service
-
+bool XrdHttpProtocol::isRequiredGridmap = false;
 int XrdHttpProtocol::sslverifydepth = 9;
+XrdSysRWLock XrdHttpProtocol::x509_store_lock;
+X509_STORE *XrdHttpProtocol::verify_store = NULL;
 SSL_CTX *XrdHttpProtocol::sslctx = 0;
 BIO *XrdHttpProtocol::sslbio_err = 0;
 XrdCryptoFactory *XrdHttpProtocol::myCryptoFactory = 0;
 XrdHttpSecXtractor *XrdHttpProtocol::secxtractor = 0;
+bool XrdHttpProtocol::isRequiredXtractor = false;
 struct XrdHttpProtocol::XrdHttpExtHandlerInfo XrdHttpProtocol::exthandler[MAX_XRDHTTPEXTHANDLERS];
 int XrdHttpProtocol::exthandlercnt = 0;
 std::map< std::string, std::string > XrdHttpProtocol::hdr2cgimap; 
@@ -116,65 +118,6 @@ XrdObjectQ<XrdHttpProtocol>
 XrdHttpProtocol::ProtStack("ProtStack",
         "xrootd protocol anchor");
 
-/******************************************************************************/
-/*                       P r o t o c o l   L o a d e r                        */
-/*                        X r d g e t P r o t o c o l                         */
-/******************************************************************************/
-
-// This protocol can live in a shared library. The interface below is used by
-// the protocol driver to obtain a copy of the protocol object that can be used
-// to decide whether or not a link is talking a particular protocol.
-//
-XrdVERSIONINFO(XrdgetProtocol, xrdhttp);
-
-extern "C" {
-
-  XrdProtocol *XrdgetProtocol(const char *pname, char *parms,
-          XrdProtocol_Config *pi) {
-    XrdProtocol *pp = 0;
-    const char *txt = "completed.";
-
-    // Put up the banner
-    //
-    pi->eDest->Say("Copr. 2012 CERN IT, an HTTP implementation for the XROOTD framework.");
-    pi->eDest->Say("++++++ HTTP protocol initialization started.");
-
-    // Return the protocol object to be used if static init succeeds
-    //
-    if (XrdHttpProtocol::Configure(parms, pi))
-      pp = (XrdProtocol *)new XrdHttpProtocol(false);
-    else txt = "failed.";
-    pi->eDest->Say("------ HTTP protocol initialization ", txt);
-    return pp;
-  }
-}
-
-/******************************************************************************/
-/*                                                                            */
-/*           P r o t o c o l   P o r t   D e t e r m i n a t i o n            */
-/*                    X r d g e t P r o t o c o l P o r t                     */
-/******************************************************************************/
-
-// This function is called early on to determine the port we need to use. The
-// default is ostensibly 1094 but can be overidden; which we allow.
-//
-XrdVERSIONINFO(XrdgetProtocolPort, xrdhttp);
-
-extern "C" {
-
-  int XrdgetProtocolPort(const char *pname, char *parms, XrdProtocol_Config *pi) {
-
-    // Figure out what port number we should return. In practice only one port
-    // number is allowed. However, we could potentially have a clustered port
-    // and several unclustered ports. So, we let this practicality slide.
-    //
-    if (pi->Port < 0) return 1094;
-    return pi->Port;
-  }
-}
-
-
-
 
 /******************************************************************************/
 /*               U g l y  O p e n S S L   w o r k a r o u n d s               */
@@ -309,128 +252,18 @@ XrdProtocol *XrdHttpProtocol::Match(XrdLink *lp) {
   else
     hp->ishttps = myishttps;
 
+  // Allocate 1MB buffer from pool
+  if (!hp->myBuff) {
+    hp->myBuff = BPool->Obtain(1024 * 1024);
+  }
+  hp->myBuffStart = hp->myBuffEnd = hp->myBuff->buff;
+
   // Bind the protocol to the link and return the protocol
   //
   hp->Link = lp;
-
-
   return (XrdProtocol *) hp;
 }
 
-
-
-
-
-
-
-
-/******************************************************************************/
-/*                               G e t V O M S D a t a                        */
-
-/******************************************************************************/
-
-
-
-int XrdHttpProtocol::GetVOMSData(XrdLink *lp) {
-  TRACEI(DEBUG, " Extracting auth info.");
-
-  X509 *peer_cert;
-
-  // No external plugin, hence we fill our XrdSec with what we can do here
-  peer_cert = SSL_get_peer_certificate(ssl);
-  TRACEI(DEBUG, " SSL_get_peer_certificate returned :" << peer_cert);
-  ERR_print_errors(sslbio_err);
-
-  if (peer_cert) {
-
-    // Add the original DN to the moninfo. Not sure if it makes sense to parametrize this or not.
-    if (SecEntity.moninfo) free(SecEntity.moninfo);
-    SecEntity.moninfo = X509_NAME_oneline(X509_get_subject_name(peer_cert), NULL, 0);
-    TRACEI(DEBUG, " Subject name is : '" << SecEntity.moninfo << "'");
-    
-    // Here we have the user DN, and try to extract an useful user name from it
-    if (SecEntity.name) free(SecEntity.name);
-    SecEntity.name = 0;
-    // To set the name we pick the first CN of the certificate subject
-    // and hope that it makes some sense, it usually does
-    char *lnpos = strstr(SecEntity.moninfo, "/CN=");
-    char bufname[64], bufname2[9];
-        
-    if (lnpos) {
-      lnpos += 4;
-      char *lnpos2 = index(lnpos, '/');
-      if (lnpos2) {
-        int l = ( lnpos2-lnpos < (int)sizeof(bufname) ? lnpos2-lnpos : (int)sizeof(bufname)-1 );
-        strncpy(bufname, lnpos, l);
-        bufname[l] = '\0';
-        
-        // Here we have the string in the buffer. Take the last 8 non-space characters
-        size_t j = 8;
-        strcpy(bufname2, "unknown-\0"); // note it's 9 chars
-        for (int i = (int)strlen(bufname)-1; i >= 0; i--) {
-          if (isalnum(bufname[i])) {
-            j--;
-            bufname2[j] = bufname[i];
-            if (j == 0) break;
-          }
-          
-        }
-        
-        SecEntity.name = strdup(bufname);
-        TRACEI(DEBUG, " Setting link name: '" << bufname2+j << "'");
-        lp->setID(bufname2+j, 0);
-      }
-    }
-    
-    if (servGMap) {
-      int e = servGMap->dn2user(SecEntity.moninfo, bufname, 127, 0);
-      if ( !e ) {
-        TRACEI(DEBUG, " Mapping Username: " << SecEntity.moninfo << " --> " << bufname);
-        if (SecEntity.name) free(SecEntity.name);
-        SecEntity.name = strdup(bufname);
-      }
-      else {
-        TRACEI(ALL, " Mapping Username: " << SecEntity.moninfo << " Failed. err: " << e);
-      }
-    }
-    
-    // If we could not find anything good, take the last 8 non-space characters of the main subject
-    if (!SecEntity.name) {
-      size_t j = 8;
-      SecEntity.name = strdup("unknown-\0"); // note it's 9 chars
-      for (int i = (int)strlen(SecEntity.moninfo)-1; i >= 0; i--) {
-        if (isalnum(SecEntity.moninfo[i])) {
-          j--;
-          SecEntity.name[j] = SecEntity.moninfo[i];
-          if (j == 0) break;
-         
-        }
-      }
-    }
-   
-    
-    
-  }
-  else return 0; // Don't fail if no cert
-
-  if (peer_cert) X509_free(peer_cert);
-
-
-
-  // Invoke our instance of the Security exctractor plugin
-  // This will fill the XrdSec thing with VOMS info, if VOMS is
-  // installed. If we have no sec extractor then do nothing, just plain https
-  // will work.
-  if (secxtractor) {
-    int r = secxtractor->GetSecData(lp, SecEntity, ssl);
-    if (r)
-      TRACEI(ALL, " Certificate data extraction failed: " << SecEntity.moninfo << " Failed. err: " << r);
-    return r;
-  }
-
-  return 0;
-}
-
 char *XrdHttpProtocol::GetClientIPStr() {
   char buf[256];
   buf[0] = '\0';
@@ -460,8 +293,8 @@ int BIO_XrdLink_write(BIO *bio, const char *data, size_t datal, size_t *written)
   BIO_clear_retry_flags(bio);
   if (ret <= 0) {
     *written = 0;
-    if ((ret == EINTR) || (ret == EINPROGRESS) || (ret == EAGAIN) || (ret == EWOULDBLOCK))
-      BIO_set_retry_read(bio);
+    if ((errno == EINTR) || (errno == EINPROGRESS) || (errno == EAGAIN) || (errno == EWOULDBLOCK))
+      BIO_set_retry_write(bio);
     return ret;
   }
   *written = ret;
@@ -480,8 +313,8 @@ int BIO_XrdLink_write(BIO *bio, const char *data, int datal)
   int ret = lp->Send(data, datal);
   BIO_clear_retry_flags(bio);
   if (ret <= 0) {
-    if ((ret == EINTR) || (ret == EINPROGRESS) || (ret == EAGAIN) || (ret == EWOULDBLOCK))
-      BIO_set_retry_read(bio);
+    if ((errno == EINTR) || (errno == EINPROGRESS) || (errno == EAGAIN) || (errno == EWOULDBLOCK))
+      BIO_set_retry_write(bio);
   }
   return ret;
 }
@@ -503,7 +336,7 @@ static int BIO_XrdLink_read(BIO *bio, char *data, size_t datal, size_t *read)
   BIO_clear_retry_flags(bio);
   if (ret <= 0) {
     *read = 0;
-    if ((ret == EINTR) || (ret == EINPROGRESS) || (ret == EAGAIN) || (ret == EWOULDBLOCK))
+    if ((errno == EINTR) || (errno == EINPROGRESS) || (errno == EAGAIN) || (errno == EWOULDBLOCK))
       BIO_set_retry_read(bio);
     return ret;
   }
@@ -522,7 +355,7 @@ static int BIO_XrdLink_read(BIO *bio, char *data, int datal)
   int ret = lp->Recv(data, datal);
   BIO_clear_retry_flags(bio);
   if (ret <= 0) {
-    if ((ret == EINTR) || (ret == EINPROGRESS) || (ret == EAGAIN) || (ret == EWOULDBLOCK))
+    if ((errno == EINTR) || (errno == EINPROGRESS) || (errno == EAGAIN) || (errno == EWOULDBLOCK))
       BIO_set_retry_read(bio);
   }
   return ret;
@@ -635,6 +468,19 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
           sbio = CreateBIO(Link);
           BIO_set_nbio(sbio, 1);
           ssl = SSL_new(sslctx);
+
+          // On newer versions of OpenSSL, we use the periodically
+          // updated store `verify_store` with the `x509_store_lock`
+          // held; older OpenSSL versions are missing `SSL_set1_*_cert_store`,
+          // so we simply rely on the global one in SSL_CTX.  The latter
+          // doesn't refresh the verify store, which causes issues with some
+          // plugins.
+#if OPENSSL_VERSION_NUMBER >= 0x10002000L
+          XrdSysRWLockHelper scopedLock(x509_store_lock, true);
+          SSL_set1_verify_cert_store(ssl, verify_store);
+          SSL_set1_chain_cert_store(ssl, verify_store);
+#endif
+
         }
 
       if (!ssl) {
@@ -674,22 +520,17 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
           ssl = 0;
           return -1;
         }
-      BIO_set_nbio(sbio, 0);
 
-      res = SSL_get_verify_result(ssl);
-      TRACEI(DEBUG, " SSL_get_verify_result returned :" << res);
+      BIO_set_nbio(sbio, 0);
       ERR_print_errors(sslbio_err);
 
-
       // Get the voms string and auth information
-      if (GetVOMSData(Link)) {
+      if (HandleAuthentication(Link)) {
           SSL_free(ssl);
           ssl = 0;
           return -1;
       }
 
-
-      if (res != X509_V_OK) return -1;
       ssldone = true;
     }
 
@@ -726,7 +567,7 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
     while ((rc = BuffgetLine(tmpline)) > 0) {
       TRACE(DEBUG, " rc:" << rc << " got hdr line: " << tmpline);
 
-      if ((rc == 2) && (tmpline[rc - 1] == '\n')) {
+      if ((rc == 2) && (tmpline.length() > 1) && (tmpline[rc - 1] == '\n')) {
         CurrentReq.headerok = true;
         TRACE(DEBUG, " rc:" << rc << " detected header end.");
         break;
@@ -738,6 +579,7 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
         int result = CurrentReq.parseFirstLine((char *)tmpline.c_str(), rc);
         if (result < 0) {
           TRACE(DEBUG, " Parsing of first line failed with " << result);
+          return -1;
         }
       }
       else
@@ -750,6 +592,18 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
 
     if (!CurrentReq.headerok) {
       TRACEI(REQ, " rc:" << rc << "Header not yet complete.");
+      
+      // Here a subtle error condition. IF we failed reading a line AND the buffer
+      // has a reasonable amount of data available THEN we consider the header
+      // as corrupted and shutdown the client
+      if ((rc <= 0) && (BuffUsed() >= 16384)) {
+        TRACEI(ALL, "Corrupted header detected, or line too long. Disconnecting client.");
+        return -1;
+      }
+        
+        
+      if (CurrentReq.reqstate > 0)
+        CurrentReq.reqstate--;
       // Waiting for more data
       return 1;
     }
@@ -826,7 +680,7 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
 
         
         CurrentReq.appendOpaque(dest, &SecEntity, hash, timenow);
-        SendSimpleResp(302, NULL, (char *) dest.c_str(), 0, 0);
+        SendSimpleResp(302, NULL, (char *) dest.c_str(), 0, 0, true);
         CurrentReq.reset();
         return -1;
       }
@@ -892,9 +746,47 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
           TRACEI(REQ, " Setting dn: " << SecEntity.moninfo);
         }
 
-        // TODO: compare the xrdhttphost with the real client IP
-        // If they are different then reject
+        nfo = CurrentReq.opaque->Get("xrdhttprole");
+        if (nfo) {
+          TRACEI(DEBUG, " Setting role: " << nfo);
+          SecEntity.role = unquote(nfo);
+          TRACEI(REQ, " Setting role: " << SecEntity.role);
+        }
 
+        nfo = CurrentReq.opaque->Get("xrdhttpgrps");
+        if (nfo) {
+          TRACEI(DEBUG, " Setting grps: " << nfo);
+          SecEntity.grps = unquote(nfo);
+          TRACEI(REQ, " Setting grps: " << SecEntity.grps);
+        }
+        
+        nfo = CurrentReq.opaque->Get("xrdhttpendorsements");
+        if (nfo) {
+          TRACEI(DEBUG, " Setting endorsements: " << nfo);
+          SecEntity.endorsements = unquote(nfo);
+          TRACEI(REQ, " Setting endorsements: " << SecEntity.endorsements);
+        }
+        
+        nfo = CurrentReq.opaque->Get("xrdhttpcredslen");
+        if (nfo) {
+          TRACEI(DEBUG, " Setting credslen: " << nfo);
+          char *s1 = unquote(nfo);
+          if (s1 && s1[0]) {
+            SecEntity.credslen = atoi(s1);
+            TRACEI(REQ, " Setting credslen: " << SecEntity.credslen);
+          }
+          if (s1) free(s1);
+        }
+        
+        if (SecEntity.credslen) {
+          nfo = CurrentReq.opaque->Get("xrdhttpcreds");
+          if (nfo) {
+            TRACEI(DEBUG, " Setting creds: " << nfo);
+            SecEntity.creds = unquote(nfo);
+            TRACEI(REQ, " Setting creds: " << SecEntity.creds);
+          }
+        }
+        
         char hash[512];
 
         calcHashes(hash, CurrentReq.resource.c_str(), (kXR_int16) CurrentReq.request,
@@ -933,9 +825,9 @@ int XrdHttpProtocol::Process(XrdLink *lp) // We ignore the argument here
   // for authorization in the paths that it manages
   if (!Bridge && !FindMatchingExtHandler(CurrentReq)) {
     if (SecEntity.name)
-      Bridge = XrdXrootd::Bridge::Login(&CurrentReq, Link, &SecEntity, SecEntity.name, "XrdHttp");
+      Bridge = XrdXrootd::Bridge::Login(&CurrentReq, Link, &SecEntity, SecEntity.name, ishttps ? "https" : "http");
     else
-      Bridge = XrdXrootd::Bridge::Login(&CurrentReq, Link, &SecEntity, "unknown", "XrdHttp");
+      Bridge = XrdXrootd::Bridge::Login(&CurrentReq, Link, &SecEntity, "unknown", ishttps ? "https" : "http");
       
     if (!Bridge) {
       TRACEI(REQ, " Authorization failed.");
@@ -1082,6 +974,7 @@ int XrdHttpProtocol::Config(const char *ConfigFN, XrdOucEnv *myEnv) {
       else if TS_Xeq("cert", xsslcert);
       else if TS_Xeq("key", xsslkey);
       else if TS_Xeq("cadir", xsslcadir);
+      else if TS_Xeq("cipherfilter", xsslcipherfilter);
       else if TS_Xeq("gridmap", xgmap);
       else if TS_Xeq("cafile", xsslcafile);
       else if TS_Xeq("secretkey", xsecretkey);
@@ -1284,8 +1177,8 @@ int XrdHttpProtocol::getDataOneShot(int blen, bool wait) {
     }
 
     if (rlen < 0) {
-      Link->setEtext("link timeout");
-      return 1;
+      Link->setEtext("link timeout or other error");
+      return -1;
     }
 
 
@@ -1377,7 +1270,7 @@ int XrdHttpProtocol::BuffgetData(int blen, char **data, bool wait) {
   
   if (wait && (blen > BuffUsed())) {
     TRACE(REQ, "BuffgetData: need to read " << blen - BuffUsed() << " bytes");
-    if (getDataOneShot(blen - BuffUsed(), true) < 0) return 0;
+    if ( getDataOneShot(blen - BuffUsed(), true) ) return 0;
   }
 
   if (myBuffStart < myBuffEnd) {
@@ -1415,7 +1308,7 @@ int XrdHttpProtocol::SendData(const char *body, int bodylen) {
   return 0;
 }
 
-int XrdHttpProtocol::StartSimpleResp(int code, const char *desc, const char *header_to_add, long long bodylen) {
+int XrdHttpProtocol::StartSimpleResp(int code, const char *desc, const char *header_to_add, long long bodylen, bool keepalive) {
   std::stringstream ss;
   const std::string crlf = "\r\n";
 
@@ -1424,14 +1317,23 @@ int XrdHttpProtocol::StartSimpleResp(int code, const char *desc, const char *hea
     ss << desc;
   } else {
     if (code == 200) ss << "OK";
-    else if (code == 206) ss << "Partial content";
+    else if (code == 100) ss << "Continue";
+    else if (code == 206) ss << "Partial Content";
     else if (code == 302) ss << "Redirect";
-    else if (code == 404) ss << "Not found";
+    else if (code == 403) ss << "Forbidden";
+    else if (code == 404) ss << "Not Found";
+    else if (code == 405) ss << "Method Not Allowed";
+    else if (code == 500) ss << "Internal Server Error";
     else ss << "Unknown";
   }
   ss << crlf;
+  if (keepalive && (code != 100))
+    ss << "Connection: Keep-Alive" << crlf;
+  else
+    ss << "Connection: Close" << crlf;
 
-  if (bodylen >= 0) ss << "Content-Length: " << bodylen << crlf;
+  if ((bodylen >= 0) && (code != 100))
+    ss << "Content-Length: " << bodylen << crlf;
 
   if (header_to_add)
     ss << header_to_add << crlf;
@@ -1446,7 +1348,7 @@ int XrdHttpProtocol::StartSimpleResp(int code, const char *desc, const char *hea
   return 0;
 }
 
-int XrdHttpProtocol::StartChunkedResp(int code, const char *desc, const char *header_to_add) {
+int XrdHttpProtocol::StartChunkedResp(int code, const char *desc, const char *header_to_add, bool keepalive) {
   const std::string crlf = "\r\n";
 
   std::stringstream ss;
@@ -1456,7 +1358,7 @@ int XrdHttpProtocol::StartChunkedResp(int code, const char *desc, const char *he
   ss << "Transfer-Encoding: chunked";
 
   TRACEI(RSP, "Starting chunked response");
-  return StartSimpleResp(code, desc, ss.str().c_str(), -1);
+  return StartSimpleResp(code, desc, ss.str().c_str(), -1, keepalive);
 }
 
 int XrdHttpProtocol::ChunkResp(const char *body, long long bodylen) {
@@ -1484,14 +1386,14 @@ int XrdHttpProtocol::ChunkResp(const char *body, long long bodylen) {
 /// Header_to_add is a set of header lines each CRLF terminated to be added to the header
 /// Returns 0 if OK
 
-int XrdHttpProtocol::SendSimpleResp(int code, const char *desc, const char *header_to_add, const char *body, long long bodylen) {
+int XrdHttpProtocol::SendSimpleResp(int code, const char *desc, const char *header_to_add, const char *body, long long bodylen, bool keepalive) {
 
   long long content_length = bodylen;
   if (bodylen <= 0) {
     content_length = body ? strlen(body) : 0;
   }
 
-  if (StartSimpleResp(code, desc, header_to_add, content_length) < 0)
+  if (StartSimpleResp(code, desc, header_to_add, content_length, keepalive) < 0)
     return -1;
 
   //
@@ -1605,11 +1507,6 @@ int XrdHttpProtocol::Configure(char *parms, XrdProtocol_Config * pi) {
 }
 
 
-
-
-
-
-
 // --------------
 // security stuff
 // --------------
@@ -1634,11 +1531,58 @@ extern "C" int verify_callback(int ok, X509_STORE_CTX * store) {
   return ok;
 }
 
-
-
-
-
 /// Initialization of the ssl security
+//  Only scheduled on sufficiently new OpenSSL.
+
+class XrdCertStoreJob : XrdJob
+{
+public:
+
+  void DoIt() {XrdHttpProtocol::PeriodicUpdate();
+               Sched->Schedule((XrdJob *)this, time(0)+iVal);
+              }
+
+  XrdCertStoreJob(XrdScheduler *schP, int iV)
+                  : XrdJob("cert store updater"),
+                       Sched(schP), iVal(iV)
+                    {Sched->Schedule((XrdJob *)this, time(0)+iVal);}
+  ~XrdCertStoreJob() {}
+private:
+
+  XrdScheduler *Sched;
+  int           iVal;
+};
+
+void XrdHttpProtocol::PeriodicUpdate() {
+
+    X509_STORE *new_store = PrepareStore();
+    TRACE(EMSG, "Updating new cert store");
+    if (new_store) {
+      XrdSysRWLockHelper scopedLock(x509_store_lock, false);
+      X509_STORE_free(verify_store);
+      verify_store = new_store;
+    }
+}
+
+X509_STORE *XrdHttpProtocol::PrepareStore() {
+  X509_STORE *store = X509_STORE_new();
+
+  if (!store) {
+    eDest.Say(" error: failed to allocate new certificate store");
+    return NULL;
+  }
+
+  X509_STORE_set_depth(store, sslverifydepth);
+  X509_STORE_set_flags(store, X509_V_FLAG_ALLOW_PROXY_CERTS);
+
+  if (!X509_STORE_load_locations(store, sslcafile, sslcadir)) {
+    TRACE(EMSG, " Error setting the ca file or directory.");
+    ERR_print_errors(sslbio_err);
+    return NULL;
+  }
+
+  return store;
+}
 
 int XrdHttpProtocol::InitSecurity() {
   
@@ -1663,7 +1607,10 @@ int XrdHttpProtocol::InitSecurity() {
 
   const SSL_METHOD *meth;
 
-#ifdef HAVE_TLS
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+  meth = TLS_method();
+  eDest.Say(" Using TLS_method");
+#elif defined (HAVE_TLS)
   meth = TLS_method();
   eDest.Say(" Using TLS");
 #elif defined (HAVE_TLS12)
@@ -1717,8 +1664,48 @@ int XrdHttpProtocol::InitSecurity() {
     }
   }
 
+  // Initialize a store for use on individual SSL objects and schedule
+  // a periodic update.  These separate stores are only usable on versions
+  // of OpenSSL 1.1.0 or later.
+#if OPENSSL_VERSION_NUMBER >= 0x10002000L
+  {
+    XrdSysRWLockHelper scopedLock(x509_store_lock, false);
+    verify_store = PrepareStore();
+    if (!verify_store) {
+      exit(1);
+    }
+  }
+  new XrdCertStoreJob(Sched, 3600);
+#endif
+
+  // Use default cipherlist filter if none is provided
+#if OPENSSL_VERSION_NUMBER >= 0x10002000L
+  // For OpenSSL v1.0.2+ (EL7+), use recommended cipher list from Mozilla
+  // https://ssl-config.mozilla.org/#config=intermediate&openssl=1.0.2k&guideline=5.4
+  if (!sslcipherfilter) sslcipherfilter = (char *) "ECDHE-ECDSA-AES128-GCM-SHA256:ECDHE-RSA-AES128-GCM-SHA256:ECDHE-ECDSA-AES256-GCM-SHA384:ECDHE-RSA-AES256-GCM-SHA384:ECDHE-ECDSA-CHACHA20-POLY1305:ECDHE-RSA-CHACHA20-POLY1305:DHE-RSA-AES128-GCM-SHA256:DHE-RSA-AES256-GCM-SHA384";
+#else
+  if (!sslcipherfilter) sslcipherfilter = (char *) "ALL:!LOW:!EXP:!MD5:!MD2";
+#endif
+  /* Apply the cipherlist filtering. */
+  if (!SSL_CTX_set_cipher_list(sslctx, sslcipherfilter)) {
+    TRACE(EMSG, " Error setting the cipherlist filter.");
+    ERR_print_errors(sslbio_err);
+    exit(1);
+  }
+
+  // NOTE(bbockelm): OpenSSL docs note that peer_chain is not available
+  // in reused sessions.  We should build a session cache, but we just
+  // disable sessions for now.  The session cache breaks X509 auth that
+  // is built-in to XrdHttp.
+  SSL_CTX_set_session_cache_mode(sslctx, SSL_SESS_CACHE_OFF);
+  SSL_CTX_set_options(sslctx, SSL_OP_NO_TICKET);
+
+#if SSL_CTRL_SET_ECDH_AUTO
+    // Enable elliptic-curve support
+    // not needed in OpenSSL 1.1.0+
+    SSL_CTX_set_ecdh_auto(sslctx, 1);
+#endif
 
-  SSL_CTX_set_cipher_list(sslctx, "ALL:!LOW:!EXP:!MD5:!MD2");
   //SSL_CTX_set_purpose(sslctx, X509_PURPOSE_ANY);
   SSL_CTX_set_mode(sslctx, SSL_MODE_AUTO_RETRY);
 
@@ -1801,10 +1788,11 @@ void XrdHttpProtocol::Reset() {
   CurrentReq.reset();
   CurrentReq.reqstate = 0;
 
-  if (!myBuff) {
-    myBuff = BPool->Obtain(1024 * 1024);
+  if (myBuff) {
+    BPool->Release(myBuff);
+    myBuff = 0;
   }
-  myBuffStart = myBuffEnd = myBuff->buff;
+  myBuffStart = myBuffEnd = 0;
 
   DoingLogin = false;
 
@@ -1951,12 +1939,13 @@ int XrdHttpProtocol::xsslkey(XrdOucStream & Config) {
 
 /* Function: xgmap
 
-   Purpose:  To parse the directive: gridmap <path>
+   Purpose:  To parse the directive: gridmap [required] <path>
 
-             <path>    the path of the gridmap file to be used. Normally
-                       it's /etc/grid-security/gridmap
-                       No mapfile means no translation required
-                       Pointing to a non existing mapfile is an error
+     required   optional parameter which if present treats any grimap errors
+                as fatal.
+     <path>     the path of the gridmap file to be used. Normally it's
+                /etc/grid-security/gridmap. No mapfile means no translation
+                required. Pointing to a non existing mapfile is an error.
 
   Output: 0 upon success or !0 upon failure.
  */
@@ -1972,11 +1961,23 @@ int XrdHttpProtocol::xgmap(XrdOucStream & Config) {
     return 1;
   }
 
+  // Handle optional parameter "required"
+  //
+  if (!strncmp(val, "required", 8)) {
+    isRequiredGridmap = true;
+    val = Config.GetWord();
+
+    if (!val || !val[0]) {
+      eDest.Emsg("Config", "HTTP X509 gridmap file missing after [required] "
+                 "parameter");
+      return 1;
+    }
+  }
+
   // Record the path
   //
   if (gridmap) free(gridmap);
   gridmap = strdup(val);
-
   return 0;
 }
 
@@ -2351,7 +2352,7 @@ int XrdHttpProtocol::xstaticpreload(XrdOucStream & Config) {
 
 
 /******************************************************************************/
-/*                          x s e l f h t t p s 2 h t t p                        */
+/*                          x s e l f h t t p s 2 h t t p                     */
 /******************************************************************************/
 
 /* Function: selfhttps2http
@@ -2390,14 +2391,17 @@ int XrdHttpProtocol::xselfhttps2http(XrdOucStream & Config) {
 
 /* Function: xsecxtractor
 
-   Purpose:  To parse the directive: secxtractor <path>
+   Purpose:  To parse the directive: secxtractor [required] <path> <params>
 
-             <path>    the path of the plugin to be loaded
+     required    optional parameter which if present treats any secxtractor
+                 errors as fatal.
+     <path>      the path of the plugin to be loaded
+     <params>    parameters passed to the secxtractor library
 
   Output: 0 upon success or !0 upon failure.
  */
 
-int XrdHttpProtocol::xsecxtractor(XrdOucStream & Config) {
+int XrdHttpProtocol::xsecxtractor(XrdOucStream& Config) {
   char *val;
 
   // Get the path
@@ -2406,26 +2410,40 @@ int XrdHttpProtocol::xsecxtractor(XrdOucStream & Config) {
   if (!val || !val[0]) {
     eDest.Emsg("Config", "No security extractor plugin specified.");
     return 1;
-  } else {
-
-      // Try to load the plugin (if available) that extracts info from the user cert/proxy
-      //
-      if (LoadSecXtractor(&eDest, val, 0))
-          return 1;
   }
 
+  // Handle optional parameter [required]
+  //
+  if (!strncmp(val, "required", 8)) {
+    isRequiredXtractor = true;
+    val = Config.GetWord();
+
+    if (!val || !val[0]) {
+      eDest.Emsg("Config", "No security extractor plugin after [required] "
+                 "parameter");
+      return 1;
+    }
+  }
+
+  char libName[4096];
+  strncpy(libName, val, sizeof(libName));
+  libName[sizeof(libName) - 1] = '\0';
+  char libParms[4096];
+
+  if (!Config.GetRest(libParms, 4095)) {
+    eDest.Emsg("Config", "secxtractor config params longer than 4k");
+    return 1;
+  }
+
+  // Try to load the plugin (if available) that extracts info from the user cert/proxy
+  //
+  if (LoadSecXtractor(&eDest, val, 0))
+    return 1;
 
   return 0;
 }
 
 
-
-
-
-
-
-
-
 /******************************************************************************/
 /*                            x e x t h a n d l e r                           */
 /******************************************************************************/
@@ -2587,6 +2605,40 @@ int XrdHttpProtocol::xsslcadir(XrdOucStream & Config) {
   return 0;
 }
 
+
+/******************************************************************************/
+/*                      x s s l c i p h e r f i l t e r                       */
+/******************************************************************************/
+
+/* Function: xsslcipherfilter
+
+   Purpose:  To parse the directive: cipherfilter <filter>
+
+             <filter>    the filter string to be used when generating
+                         the SSL cipher list
+
+   Output: 0 upon success or !0 upon failure.
+ */
+
+int XrdHttpProtocol::xsslcipherfilter(XrdOucStream & Config) {
+  char *val;
+
+  // Get the filter string
+  //
+  val = Config.GetWord();
+  if (!val || !val[0]) {
+    eDest.Emsg("Config", "SSL cipherlist filter string not specified");
+    return 1;
+  }
+
+  // Record the filter string
+  //
+  if (sslcipherfilter) free(sslcipherfilter);
+  sslcipherfilter = strdup(val);
+
+  return 0;
+}
+
 /******************************************************************************/
 /*                                x t r a c e                                 */
 /******************************************************************************/
@@ -2596,7 +2648,7 @@ int XrdHttpProtocol::xsslcadir(XrdOucStream & Config) {
    Purpose:  To parse the directive: trace <events>
 
              <events> the blank separated list of events to trace. Trace
-                      directives are cummalative.
+                      directives are cumulative.
 
    Output: 0 upon success or 1 upon failure.
  */
@@ -2660,6 +2712,7 @@ int XrdHttpProtocol::doStat(char *fname) {
   l = strlen(fname) + 1;
   CurrentReq.xrdreq.stat.dlen = htonl(l);
 
+  if (!Bridge) return -1;
   b = Bridge->Run((char *) &CurrentReq.xrdreq, fname, l);
   if (!b) {
     return -1;
@@ -2670,7 +2723,24 @@ int XrdHttpProtocol::doStat(char *fname) {
 }
 
 
+int XrdHttpProtocol::doChksum(const XrdOucString &fname) {
+  size_t length;
+  memset(&CurrentReq.xrdreq, 0, sizeof (ClientRequest));
+  CurrentReq.xrdreq.query.requestid = htons(kXR_query);
+  CurrentReq.xrdreq.query.infotype = htons(kXR_Qcksum);
+  memset(CurrentReq.xrdreq.query.reserved1, '\0', sizeof(CurrentReq.xrdreq.query.reserved1));
+  memset(CurrentReq.xrdreq.query.fhandle, '\0', sizeof(CurrentReq.xrdreq.query.fhandle));
+  memset(CurrentReq.xrdreq.query.reserved2, '\0', sizeof(CurrentReq.xrdreq.query.reserved2));
+  length = fname.length() + 1;
+  CurrentReq.xrdreq.query.dlen = htonl(length);
 
+  if (!Bridge) return -1;
+
+  return Bridge->Run(reinterpret_cast<char *>(&CurrentReq.xrdreq), const_cast<char *>(fname.c_str()), length) ? 0 : -1;
+}
+
+
+static XrdVERSIONINFODEF(compiledVer, XrdHttpProtocolTest, XrdVNUMBER, XrdVERSION);
 
 // Loads the SecXtractor plugin, if available
 int XrdHttpProtocol::LoadSecXtractor(XrdSysError *myeDest, const char *libName,
@@ -2680,8 +2750,7 @@ int XrdHttpProtocol::LoadSecXtractor(XrdSysError *myeDest, const char *libName,
   // We don't want to load it more than once
   if (secxtractor) return 1;
   
-    XrdVersionInfo *myVer = &XrdVERSIONINFOVAR(XrdgetProtocol);
-    XrdOucPinLoader myLib(myeDest, myVer, "secxtractorlib", libName);
+    XrdOucPinLoader myLib(myeDest, &compiledVer, "secxtractorlib", libName);
     XrdHttpSecXtractor *(*ep)(XrdHttpSecXtractorArgs);
 
     // Get the entry point of the object creator
@@ -2709,8 +2778,7 @@ int XrdHttpProtocol::LoadExtHandler(XrdSysError *myeDest, const char *libName,
     return 1;
   }
   
-  XrdVersionInfo *myVer = &XrdVERSIONINFOVAR(XrdgetProtocol);
-  XrdOucPinLoader myLib(myeDest, myVer, "exthandlerlib", libName);
+  XrdOucPinLoader myLib(myeDest, &compiledVer, "exthandlerlib", libName);
   XrdHttpExtHandler *(*ep)(XrdHttpExtHandlerArgs);
   
   // Get the entry point of the object creator
