@@ -30,6 +30,7 @@
 
 #include <errno.h>
 #include <fcntl.h>
+#include <stdio.h>
 #include <sys/time.h>
 #include <sys/param.h>
 #include <sys/resource.h>
@@ -44,6 +45,7 @@
 #include "XrdPosix/XrdPosixTrace.hh"
 #include "XrdPosix/XrdPosixXrootdPath.hh"
 
+#include "XrdSys/XrdSysError.hh"
 #include "XrdSys/XrdSysTimer.hh"
 
 /******************************************************************************/
@@ -54,6 +56,7 @@ namespace XrdPosixGlobals
 {
 extern XrdOucCache2    *theCache;
 extern XrdOucName2Name *theN2N;
+extern XrdSysError     *eDest;
 extern int              ddInterval;
 extern int              ddMaxTries;
        int              ddNumLost = 0;
@@ -151,6 +154,7 @@ void* XrdPosixFile::DelayedDestroy(void* vpf)
    EPNAME("DDestroy");
 
    XrdCl::XRootDStatus Status;
+   std::string statusMsg;
    const char *eTxt;
    XrdPosixFile *fCurr, *fNext;
    int ddCount;
@@ -185,13 +189,23 @@ do{if (doWait)
         {fNext = fCurr->nextFile;
          if (!((ioActive = fCurr->XCio->ioActive())) && !fCurr->Refs())
             {if (fCurr->Close(Status)) {delete fCurr; ddCount--; continue;}
-                else eTxt = Status.ToString().c_str();
-            } else   eTxt = (ioActive ? "active I/O" : "callback");
+                else {statusMsg = Status.ToString();
+                      eTxt = statusMsg.c_str();
+                     }
+            } else    eTxt = (ioActive ? "active I/O" : "callback");
 
          if (fCurr->numTries > XrdPosixGlobals::ddMaxTries)
             {XrdPosixGlobals::ddNumLost++; ddCount--;
-             DMSG("DDestroy", eTxt <<" timeout closing " <<fCurr->Origin()
+             if (XrdPosixGlobals::eDest)
+                {char buff[256];
+                 snprintf(buff, sizeof(buff), "(%d) %s",
+                                XrdPosixGlobals:: ddNumLost, eTxt);
+                 XrdPosixGlobals::eDest->Emsg("DDestroy",
+                                         "timeout closing", fCurr->Origin());
+                } else {
+                 DMSG("DDestroy", eTxt <<" timeout closing " <<fCurr->Origin()
                         <<' ' <<XrdPosixGlobals::ddNumLost <<" objects lost");
+                }
              fCurr->nextFile = ddLost;
              ddLost = fCurr;
              fCurr->Close(Status);
@@ -243,6 +257,11 @@ void XrdPosixFile::DelayedDestroy(XrdPosixFile *fp)
 
 bool XrdPosixFile::Close(XrdCl::XRootDStatus &Status)
 {
+// If this is a defered open, disable any future calls as we are ready to
+// shutdown this beast!
+//
+   if (PrepIO) PrepIO->Disable();
+
 // If we don't need to close the file, then return success. Otherwise, do the
 // actual close and return the status. We should have already been removed
 // from the file table at this point and should be unlocked.
@@ -309,6 +328,8 @@ int XrdPosixFile::Fstat(struct stat &buf)
 /******************************************************************************/
 /*                        H a n d l e R e s p o n s e                         */
 /******************************************************************************/
+
+// Note: This response handler is only used for async open requests!
   
 void XrdPosixFile::HandleResponse(XrdCl::XRootDStatus *status,
                                   XrdCl::AnyObject    *response)
@@ -322,7 +343,8 @@ void XrdPosixFile::HandleResponse(XrdCl::XRootDStatus *status,
    if (!(status->IsOK()))          rc = XrdPosixMap::Result(*status);
       else if (!Finalize(&Status)) rc = XrdPosixMap::Result(Status);
 
-// Issue callback with the correct result
+// Issue XrdPosixCallBack callback with the correct result. Note: errors are
+// indicated with result set to -1 and errno set to the error number.
 //
    xeqCB->Complete(rc);
 
@@ -330,7 +352,7 @@ void XrdPosixFile::HandleResponse(XrdCl::XRootDStatus *status,
 //
    delete status;
    delete response;
-   if (rc) delete this;
+   if (rc < 0) delete this;
 }
 
 /******************************************************************************/
@@ -366,11 +388,13 @@ int XrdPosixFile::Read (char *Buff, long long Offs, int Len)
    XrdCl::XRootDStatus Status;
    uint32_t bytes;
 
-// Issue read and return appropriately
+// Issue read and return appropriately.
 //
+   Ref();
    Status = clFile.Read((uint64_t)Offs, (uint32_t)Len, Buff, bytes);
+   unRef();
 
-   return (Status.IsOK() ? (int)bytes : XrdPosixMap::Result(Status));
+   return (Status.IsOK() ? (int)bytes : XrdPosixMap::Result(Status, false));
 }
   
 /******************************************************************************/
@@ -384,11 +408,15 @@ void XrdPosixFile::Read (XrdOucCacheIOCB &iocb, char *buff, long long offs,
 
 // Issue read
 //
+   Ref();
    Status = clFile.Read((uint64_t)offs, (uint32_t)rlen, buff, rhp);
 
 // Check status
 //
-   if (!Status.IsOK()) rhp->Sched(-XrdPosixMap::Result(Status));
+   if (!Status.IsOK())
+      {rhp->Sched(XrdPosixMap::Result(Status, false));
+       unRef();
+      }
 }
 
 /******************************************************************************/
@@ -416,12 +444,14 @@ int XrdPosixFile::ReadV (const XrdOucIOVec *readV, int n)
 // Issue the readv. We immediately delete the vrInfo as w don't need it as a
 // readv will succeed only if actually read the number of bytes requested.
 //
+   Ref();
    Status = clFile.VectorRead(chunkVec, (void *)0, vrInfo);
+   unRef();
    delete vrInfo;
 
 // Return appropriate result
 //
-   return (Status.IsOK() ? nbytes : XrdPosixMap::Result(Status));
+   return (Status.IsOK() ? nbytes : XrdPosixMap::Result(Status, false));
 }
 
 /******************************************************************************/
@@ -447,11 +477,15 @@ void XrdPosixFile::ReadV(XrdOucCacheIOCB &iocb, const XrdOucIOVec *readV, int n)
 //
    XrdPosixFileRH *rhp =  XrdPosixFileRH::Alloc(&iocb, this, 0, nbytes,
                                                 XrdPosixFileRH::isReadV);
+   Ref();
    Status = clFile.VectorRead(chunkVec, (void *)0, rhp);
 
 // Return appropriate result
 //
-   if (!Status.IsOK()) rhp->Sched(-XrdPosixMap::Result(Status));
+   if (!Status.IsOK())
+      {rhp->Sched(XrdPosixMap::Result(Status, false));
+       unRef();
+      }
 }
 
 /******************************************************************************/
@@ -464,9 +498,11 @@ bool XrdPosixFile::Stat(XrdCl::XRootDStatus &Status, bool force)
 
 // Get the stat information from the open file
 //
+   Ref();
    Status = clFile.Stat(force, sInfo);
    if (!Status.IsOK())
-      {delete sInfo;
+      {unRef();
+       delete sInfo;
        return false;
       }
 
@@ -479,6 +515,7 @@ bool XrdPosixFile::Stat(XrdCl::XRootDStatus &Status, bool force)
 
 // Delete our status information and return final result
 //
+   unRef();
    delete sInfo;
    return true;
 }
@@ -487,6 +524,23 @@ bool XrdPosixFile::Stat(XrdCl::XRootDStatus &Status, bool force)
 /*                                  S y n c                                   */
 /******************************************************************************/
 
+int XrdPosixFile::Sync()
+{
+   XrdCl::XRootDStatus Status;
+
+// Issue the Sync
+//
+   Ref();
+   Status = clFile.Sync();
+   unRef();
+
+// Return result
+//
+   return XrdPosixMap::Result(Status, false);
+}
+
+/******************************************************************************/
+
 void XrdPosixFile::Sync(XrdOucCacheIOCB &iocb)
 {
    XrdCl::XRootDStatus Status;
@@ -499,7 +553,26 @@ void XrdPosixFile::Sync(XrdOucCacheIOCB &iocb)
 
 // Check status
 //
-   if (!Status.IsOK()) rhp->Sched(-XrdPosixMap::Result(Status));
+   if (!Status.IsOK()) rhp->Sched(XrdPosixMap::Result(Status, false));
+}
+  
+/******************************************************************************/
+/*                                 T r u n c                                  */
+/******************************************************************************/
+
+int XrdPosixFile::Trunc(long long Offset)
+{
+   XrdCl::XRootDStatus Status;
+
+// Issue truncate request
+//
+   Ref();
+   Status = clFile.Truncate((uint64_t)Offset);
+   unRef();
+
+// Return results
+//
+   return XrdPosixMap::Result(Status);
 }
   
 /******************************************************************************/
@@ -512,7 +585,9 @@ int XrdPosixFile::Write(char *Buff, long long Offs, int Len)
 
 // Issue read and return appropriately
 //
+   Ref();
    Status = clFile.Write((uint64_t)Offs, (uint32_t)Len, Buff);
+   unRef();
 
    return (Status.IsOK() ? Len : XrdPosixMap::Result(Status));
 }
@@ -528,20 +603,13 @@ void XrdPosixFile::Write(XrdOucCacheIOCB &iocb, char *buff, long long offs,
 
 // Issue read
 //
+   Ref();
    Status = clFile.Write((uint64_t)offs, (uint32_t)wlen, buff, rhp);
 
 // Check status
 //
-   if (!Status.IsOK()) rhp->Sched(-XrdPosixMap::Result(Status));
-}
-  
-/******************************************************************************/
-/*                                  D o I t                                   */
-/******************************************************************************/
-void XrdPosixFile::DoIt()
-{
-// Virtual function of XrdJob.
-// Called from XrdPosixXrootd::Close if the file is still IO active.
-
-   delete this;
+   if (!Status.IsOK())
+      {rhp->Sched(XrdPosixMap::Result(Status));
+       unRef();
+      }
 }
